///////////////////////////////////////////////////////////////////////////
//
// File: model_gps.c
// Author: Michael Janssen
// Date: 20 July 2005
//
///////////////////////////////////////////////////////////////////////////

//#define DEBUG

#include <assert.h>
#include <math.h>
#include "stage_internal.h"
#include "gui.h"

extern stg_rtk_fig_t* fig_debug_rays;

const double STG_GPS_WATTS = 10.0;

#define STG_DEFAULT_GPS_LATBASE 0.0 
#define STG_DEFAULT_GPS_LONBASE 0.0 
#define STG_DEFAULT_GPS_ERROR 3.0


/** @defgroup model_gps GPS model
The model simulates a GPS with optional sources of error

<h2>Worldfile properties</h2>

@par Summary and default values

@verbatim
gps
(
  # gps properties
  lat_base 0.0
  lon_base 0.0
  error 3.0

  # model properties
  size [0 0]
)
@endverbatim

@par Details
- lat_base float
  - the latitude base, in degrees, default 0.0
- lon_base float
  - the longitude base, in degrees, default 0.0
- error
  - the amount of error in meters, default 3.0
*/


int gps_update( stg_model_t* mod );
int gps_startup( stg_model_t* mod );
int gps_shutdown( stg_model_t* mod );
void gps_load( stg_model_t* mod );
int gps_render_data( stg_model_t* mod, void* userp );
int gps_unrender_data( stg_model_t* mod, void* userp );
//TODO: Implement these
//int gps_render_cfg( stg_model_t* mod, void* userp );
//int gps_undender_cfg( std_model_t* mod, void* userp );

int gps_init( stg_model_t* mod ) {
  PRINT_DEBUG(" gps init "); 
  // override the default methods
  mod->f_startup = gps_startup;
  mod->f_shutdown = gps_shutdown;
  mod->f_update = NULL; // installed at startup/shutdown
  mod->f_load = gps_load;

  // Set up sensible defaults
  stg_geom_t geom;
  memset( &geom, 0, sizeof(geom));
  stg_model_set_geom( mod, &geom );  

  stg_color_t color = stg_lookup_color( "magenta" );
  stg_model_set_color( mod, color );

	stg_gps_t data;
	memset( &data, 0, sizeof(data));

  // no body, no polygons 
  stg_model_set_polygons( mod, NULL, 0); 
  
  // default parameters
  stg_gps_config_t cfg; 
  cfg.lon_base = STG_DEFAULT_GPS_LONBASE;
  cfg.lat_base = STG_DEFAULT_GPS_LATBASE;
  cfg.error = STG_DEFAULT_GPS_ERROR;

  stg_model_set_cfg( mod, &cfg, sizeof(cfg) );

  // Menus
  stg_model_add_property_toggles( mod, &mod->data, 
      gps_render_data, 
      NULL, 
      gps_unrender_data,
      NULL, 
      "gpsdata", 
      "gps data",
      TRUE );

  return 0;
}

void gps_load( stg_model_t* mod )
{
  stg_gps_config_t* now = (stg_gps_config_t*) mod->cfg; 
  assert(now); 

  stg_gps_config_t cfg;
  memcpy( &cfg, now, sizeof(cfg) );
  
  cfg.lat_base = wf_read_float(mod->id, "lat_base", now->lat_base );
  cfg.lon_base = wf_read_float(mod->id, "lon_base", now->lon_base );
  cfg.error = wf_read_float(mod->id, "error", now->error );
  
  stg_model_set_cfg(mod, &cfg, sizeof(stg_gps_config_t)); 
}

int gps_startup( stg_model_t* mod )
{
  PRINT_DEBUG( "gps startup" );  
  
  mod->f_update = gps_update;
 
  stg_model_set_watts( mod, STG_GPS_WATTS );
  return 0;
}

int gps_shutdown( stg_model_t* mod )
{
  mod->f_update = NULL;

  stg_model_set_watts( mod, 0 );

  // this will unrender the data
	stg_model_set_data(mod, NULL, 0); 
  
  return 0;
}

///////////////////////////////////////////////////////////////////////////
// GPS Update!

double mToLat(stg_meters_t m) { 
  return (m / (2.0 * M_PI * 6356752.3142 / 360.0)); 
} 

double mToLon(stg_meters_t m) { 
  return (m / (2.0 * M_PI * 6378137 / 360.00)); 
}

double degError(double deg) { 
  double percentErr; 

  srand(time(0)); // TODO: only do this once
  percentErr = (double) rand() / RAND_MAX; 
  return deg * percentErr; 
}

int gps_update( stg_model_t* mod )
{
  PRINT_DEBUG( "gps update" );

  if( mod->subs < 1 )
    return 0;

	stg_pose_t pos; 

	stg_model_get_global_pose(mod, &pos);

	stg_gps_config_t *cfg = (stg_gps_config_t*) mod->cfg;

  stg_gps_t data;

  struct timeval curtime;
  gettimeofday(&curtime, NULL);

  data.sec = curtime.tv_sec;
  data.usec = curtime.tv_usec;
	// NOTE: Units are deg 
  data.latitude =  mToLat(pos.y) + cfg->lat_base;
  data.longitude = mToLon(pos.x) + cfg->lon_base;
  data.latitude = data.latitude + degError(mToLat(cfg->error));
  data.longitude = data.longitude + degError(mToLon(cfg->error));


  // PRINT_DEBUG4("(%f, %f) -> gps lat: %f lon: %f", pos.x, pos.y, data.latitude, data.longitude); 
	stg_model_set_data(mod, &data, sizeof(stg_gps_t)); 

  return 0;
}

int gps_render_data( stg_model_t* mod, void* userp )
{
  stg_rtk_fig_t* fig = stg_model_get_fig( mod, "gps_data_fig"); 

  if (!fig)
    fig = stg_model_fig_create( mod, "gps_data_fig", "top", STG_LAYER_NEIGHBORDATA ); 

  stg_rtk_fig_clear( fig ); 

  stg_rtk_fig_color_rgb32( fig, stg_lookup_color( STG_GENERIC_COLOR ) );

  stg_gps_t *gpsd = (stg_gps_t*) mod->data; 
  if (mod->subs > 0) { 
    char buf[256];
    snprintf( buf, 255, "LAT: %.6f deg\nLON: %.6f deg", gpsd->latitude, gpsd->longitude);
    stg_rtk_fig_text( fig, 0.4, 0.2, 0, buf); 
  } 
  return 0; 
}

int gps_unrender_data( stg_model_t* mod, void* userp)
{
  stg_rtk_fig_clear( stg_model_get_fig( mod, "gps_data_fig" ) ); 
  return 1;
}
